#include <stdlib.h>
#include <stdio.h>

struct matrix_s {
    int m;
    int n;
    float** values;
};

enum matrix_operation_e {
    ADD,
    MULT,
};

struct matrix_s* matrix_create(const int m, const int n)
{
    if (n <= 0 || m <= 0)
    {
        return NULL;
    }

    struct matrix_s* matrix = malloc(sizeof(struct matrix_s));

    if (matrix == NULL)
    {
        return NULL;
    }

    matrix->m = m;
    matrix->n = n;
    matrix->values = malloc(m*sizeof(float*));
    if (matrix->values == NULL)
    {
        free(matrix);
        return NULL;
    }

    for (int i = 0; i < m; i++)
    {
        matrix->values[i] = malloc(n*sizeof(float));
        if (matrix->values[i] == NULL)
        {
            for (int j = 0; j < n; j++)
            {
                free(matrix->values[j]);
            }
            free(matrix->values);
            free(matrix);
            return NULL;
        }
    }

    return matrix;
}

void matrix_destroy(struct matrix_s* const matrix)
{
    for (int i = 0; i < matrix->m; i++)
    {
        free(matrix->values[i]);
    }
    free(matrix->values);
    free(matrix);
}

int matrix_set_value(const struct matrix_s* const matrix, const int i, const int j, const float value)
{
    if (matrix == NULL || i < 0 || i >= matrix->m || j < 0 || j >= matrix->n)
    {
        return 0;
    }

    matrix->values[i][j] = value;

    return 1;
}

int matrix_get_value(const struct matrix_s* const matrix, const int i, const int j, float* value)
{
    if (matrix == NULL || i < 0 || i >= matrix->m || j < 0 || j >= matrix->n || value == NULL)
    {
        return 0;
    }

    *value = matrix->values[i][j];

    return 1;
}

void matrix_print(const struct matrix_s* const matrix)
{
    if (matrix == NULL)
    {
        printf("NULL ptr\n");
        return;
    }

    for (int i = 0; i < matrix->m; i++)
    {
        for (int j = 0; j < matrix->n; j++)
        {
            float value = 0.0;
            matrix_get_value(matrix, i, j, &value);
            printf("%.1f\t", value);
        }
        printf("\n");
    }
}

struct matrix_s* matrix_operation2(const struct matrix_s* const matrix_a, const struct matrix_s* const matrix_b, const enum matrix_operation_e operation)
{
    if (matrix_a == NULL || matrix_b == NULL)
    {
        return NULL;
    }

    float a = 0, b = 0, c;
    struct matrix_s* matrix_result;

    switch (operation)
    {
        case ADD:
            if (matrix_a->n != matrix_b->n || matrix_a->m != matrix_b->m)
            {
                /* Matrix sizes mismatch */
                return NULL;
            }

            matrix_result = matrix_create(matrix_a->m, matrix_a->n);
            if (matrix_result == NULL)
            {
                return NULL;
            }

            for (int i = 0; i < matrix_result->m; i++)
            {
                for (int j = 0; j < matrix_result->n; j++)
                {
                    matrix_get_value(matrix_a, i, j, &a);
                    matrix_get_value(matrix_b, i, j, &b);

                    matrix_set_value(matrix_result, i, j, a+b);
                }
            }

            return matrix_result;
        case MULT:
            if (matrix_a->n != matrix_b->m)
            {
                /* Matrix sizes mismatch */
                return NULL;
            }

            matrix_result = matrix_create(matrix_a->m, matrix_b->n);
            if (matrix_result == NULL)
            {
                return NULL;
            }

            for (int i = 0; i < matrix_result->m; i++)
            {
                for (int j = 0; j < matrix_result->n; j++)
                {
                    c = 0;
                    for (int k = 0; k < matrix_a->n; k++)
                    {
                        matrix_get_value(matrix_a, i, k, &a);
                        matrix_get_value(matrix_b, k, j, &b);

                        c += a * b;
                    }

                    matrix_set_value(matrix_result, i, j, c);
                }
            }

            return matrix_result;
        default:
            printf("Unknown operation\n");
            return NULL;
    }
}

int main()
{
    struct matrix_s* m1 = matrix_create(2, 3);
    struct matrix_s* m2 = matrix_create(3, 2);
    struct matrix_s* m3 = NULL;

    for (int i = 0; i < 6; i++)
    {
        matrix_set_value(m1, i / 3, i % 3, (float) i);
        matrix_set_value(m2, i / 2, i % 2, (float) i+6);
    }
    printf("m1 = \n");
    matrix_print(m1);
    printf("m2 = \n");
    matrix_print(m2);

    m3 = matrix_operation2(m1, m2, MULT);

    printf("m1 * m2 = \n");
    matrix_print(m3);

    matrix_destroy(m1);
    matrix_destroy(m2);
    matrix_destroy(m3);
    return EXIT_SUCCESS;
}
